ad_proc km_db_1row_to_list  { statement_name sql args } {

    Performs the SQL query $sql. Returns a list of the results.

} {
    ad_arg_parser { bind } $args

    # Query Dispatcher (OpenACS - ben)
    set full_statement_name [db_qd_get_fullname $statement_name]

    db_with_handle db {
        set selection [db_exec 0or1row $db $full_statement_name $sql]
    }

    if { [empty_string_p $selection] } {
        return ""
    }

    set list_to_return ""

    for { set i 0 } { $i < [ns_set size $selection] } { incr i } {
        lappend list_to_return [ns_set value $selection $i]
    }


    for { set i 0 } { $i < [ns_set size $selection] } { incr i } {
        upvar 1 [ns_set key $selection $i] value
        set value [ns_set value $selection $i]
    }
    
    return $list_to_return
}

ad_proc km_db_to_named_list_list { statement_name sql args} {
    Returns a list of Tcl lists, with each sublist containing the columns returned
    by the database; if no rows are returned by the database, returns the
    empty list (empty string in Tcl 7.x and 8.x)" 
} {
    ad_arg_parser { bind } $args

    # Query Dispatcher (OpenACS )
    set full_statement_name [db_qd_get_fullname $statement_name]

    db_with_handle db {
	set selection [db_exec select $db $full_statement_name $sql]

        set keys [list]
        for {set i 0} {$i<[ns_set size $selection]} {incr i} {
            lappend keys [ns_set key $selection $i]
        }
        set list_to_return [list $keys]
        while {[ns_db getrow $db $selection]} {
            set row_list ""
            set size [ns_set size $selection]
            set i 0
            while {$i<$size} {
                lappend row_list [ns_set value $selection $i]
                incr i
            }
            lappend list_to_return $row_list
        }
    }

    return $list_to_return
}

ad_proc km_new_acs_object_id {} {
    Returns a new acs_object_id
} {
    set result [db_string km_new_acs_object_id "select acs_object_id_seq.nextval from dual"]
    return $result
}

ad_proc km_sort_key_list { object_type_id } {

    Returns a list of sort_keys available for this object_type
    (sort_keys are offered as increments of 10 and should *NEVER* be
    the same for 2 questions of one object_type)

} {
    set result ""
    set existing_keys [db_list km_sort_key_list "select sort_key from
                         sn_question_object_type_map
                         where object_type_id=$object_type_id"]
    set i 0
    while { $i < 400 } {

	set i [expr $i + 10]
	if {[elem_p $i $existing_keys]} {
	    continue
	} else {
	    lappend result $i
	}
    }
    return $result
}

ad_proc km_adt_list { { -pretty_names_p 0 } } {
    Returns a list of abstract data types as a list or as a list of tuples
    with the pretty name as the second element.
} {
    set adts {
	text option category other_category integer date file composite 
	object_link user_link content_link child_object nephew_object
	other_category
    }
    set pretty_names {
	Text "Multiple Choice" Category "Category with 'Other' field" Number Date File
	Composite "Object-Object Link" "Object-User Link" "Object-Content Link"
	"Child Object Type" "Nephew Object Type"
	
    }
 
    if { $pretty_names_p } {
	return [zip $adts $pretty_names]
    } else {
	return $adts
    }
}

ad_proc km_pretty_adt { abstract_data_type } {
    Returns the pretty abstract data type
} {
    switch $abstract_data_type {
	"text" { return Text }
	"option" { return "Multiple Choice" }
	"category" { return "Category" }
	"other_category" { return "Category with 'Other' field" }
	"date" { return "Date" }
	"file" { return "Attachment" }
	"composite" { return "Composite" }
	"object_link" { return "Object Link" }
	"user_link" { return "User Link" }
	"content_link" { return "Content Link" }
	"child_object" { return "Child Object Type" }
	"nephew_object" { return "Nephew Object Type" }
	"integer" { return "Number" }
	default { return}
    }
}


ad_proc km_pretty_tag { presentation_type } {
    Returns the pretty tag name 
} {
    switch $presentation_type {
	"shorttext" { return "Short Text" }
	"text" { return "Short Text" }
	"textarea" {return "Textarea" }
	"textarea_with_refs" {return "Textarea with Reference Buttons"}
	"select" { return Select}
	"selectmultiple" {return "Select Multiple" }
	"checkbox" {return "Checkbox"}
	"radio" {return "Radio Buttons"}
	"custom" { return "Custom"}
	"application" { return "Application Form"}
	default { return }
    }
}


