# /tcl/km-display-procs.tcl
#
# Display of data.
#
# $Id: km-display-procs.tcl,v 1.2 2009/02/13 20:40:39 jeffd Exp $

# ********** Display procs for questions of various abstract data types- No Form **********
ad_proc km_display_category_value {category_value_list} {
    Formats a list of category_id and category value tuples for simple display.
} {
    set category_names [snd [transpose $category_value_list]]
    set formatted_values [join $category_names ", "]
    return $formatted_values
}

ad_proc km_display_other_category_value {category_value_list object_id question_id} {
    Formats a list of category_id and category value tuples for simple display.
} {
    set other_value [snd [fst [km_get_object_content -question_ids [list $question_id] $object_id]]]
    set category_names [snd [transpose $category_value_list]]
    set category_list [list]
    foreach category $category_names {
	if { [string equal [string toupper $category] "OTHER"] && ![empty_string_p $other_value] } {
	    append category " ($other_value)"
	}
	lappend category_list $category
    }
    set formatted_values [join $category_list ", "]
    return $formatted_values
}

ad_proc km_display_options {answer_option_list} {
    Displays tuples in an option list for simple display.
} {
    return [join [snd [transpose $answer_option_list]] ", "]
}

# *********Buttons**********

ad_proc km_get_button_names {} {
    Returns a list of all the constant submit button names in the km library.
} {
 
    return {
       proceed
       finish
       details
       add_web_ref
       add_sn_ref 
       add_user_ref 
       add_content_ref 
       next_question 
       last_question 
       link
       browse
       all_questions
       unanswered 
       answered 
       add_child
       add_nephew
       copy_child
       copy_nephew
       volunteer
       previous_question
       next_question
       delete
    }
}

ad_proc km8_button_panel {question_id object_id object_type_id} {

    This proc returns the submit button-panel.

    Note that I'm getting the buttons to the same size manually, since netscape 4x
    can't do width as a style for an input button. You have to pass the object_id
    since the active branch path can only be determined from object data.

 } {
     set captions {}
     set button_values ""
     set next_value ""
     set next_name ""

     # If this question leads to a branch, we set the default answer as next question.
     set root_branch_p [km_root_branch_p $question_id]

     # If this question is a branch, then we need to know its parent.
     set branch_parent [km_branch_parent $question_id $object_type_id]

     # Determine the next question.
     if { $root_branch_p } {
	 # The branch continues - so we don't know where to go to yet.
	 set next_question 0
     } elseif ![empty_string_p $branch_parent] {
	 # End of branch. We determine the next question relative to the branch root.
	 set branch_root [km_get_root_branch $question_id $object_type_id]
	 set next_question [km_get_question -next_p 1 -object_id $object_id $branch_root $object_type_id]
     } else {
	 # This is no branch. So just use the current question_id for determining
	 # the next.
	 set next_question [km_get_question -next_p 1 -object_id $object_id $question_id $object_type_id]
     } 

     # Determine the previous question.
     if ![empty_string_p $branch_parent] {
	 # When in a branch, Previous always takes us back to our parent.
	 set previous_question $branch_parent 	 
     } else {
	 set previous_question [km_get_question -previous_p 1 -object_id $object_id $question_id $object_type_id]
	 if ![empty_string_p $previous_question] {
	     if [km_root_branch_p [fst $previous_question]] {
		 # The previous question is the root of a branch. To find the actual "previous"
		 # question, we have to traverse this branch until the last active question.
		 set active_branches [km_active_branch_path [fst $previous_question] $object_id $object_type_id]
		 set previous_question [fst [reverse $active_branches]]
	     }
	 }
     }

     set button_values {"&nbsp;&lt;&lt;&nbsp;" "&nbsp;&gt;&gt;&nbsp;"}
     if ![empty_string_p $previous_question] {
         set previous_value [fst $button_values]
	 set previous_name  "previous_question.[fst $previous_question]"
     }
     if ![empty_string_p $next_question] {
         set next_value [snd $button_values]
	 set next_name  "next_question.[fst $next_question]"
     }

     set button_1 ""
     set button_2 ""
     if { ![empty_string_p $previous_question] && ![empty_string_p $next_question] } {
	 set button_1 $previous_name
	 set button_2 $next_name
     } elseif ![empty_string_p $previous_question] {
	 set button_1 $previous_name
     } elseif ![empty_string_p $next_question] {
	 set button_2 $next_name
     }
     
     set button_names {"all_questions." "unanswered." "answered."}
     set button_values {"All Questions" "Completed Content"}
     return [list $button_1 \
             [fst $button_names] \
             [snd $button_names] \
             [thd $button_names] \
             $button_2]
}

# ********** Basic Display Procs **********
ad_proc km_count_object_view { user_id object_id } {
    Record an object view.
} {                                            
    if {![km_original_author_p $user_id $object_id]} {
	db_dml km_count_object_view_10 "
	    update sn_access_counts 
	    set access_count = access_count + 1
	    where object_id=:object_id
		and access_date = (trunc(sysdate,'month'))"

        if {![db_resultrows]} {
	    db_dml insert_access_count {
		insert into sn_access_counts (object_id, access_count, access_date) 
		select :object_id, 1, (trunc(sysdate,'month')) from dual
		where not exists (select 1 from sn_access_counts 
				  where object_id=:object_id 
				      and access_date = (trunc(sysdate,'month')))
	    }
	    db_dml reset_sn_object_monthly_count {
		update sn_objects
		set access_total = access_total +1,
		    access_month = 1
		where object_id = :object_id
	    }
        } else {
	    db_dml increment_sn_object_count {
		update sn_objects
		set access_total = access_total +1,
		    access_month = access_month +1
		where object_id = :object_id
	    }
	}
    }
}

ad_proc km_get_access_count { object_id } {
} {
    return [db_string access_count {
        select nvl(sum(access_count),0)
        from sn_access_counts 
        where object_id = :object_id
    }]
}

ad_proc km_get_access_count_month { object_id } {
} {
    return [db_string access_count_month {
        select nvl(sum(access_count),0)
        from sn_access_counts 
        where object_id = :object_id 
        and access_date = (trunc(sysdate,'month'))
    }]
}


ad_proc km_object_access_toolbar {object_id {category_id ""}} {
    Toolbar: [Private] Make public | Access Control | Delete Object 
} {
    set user_id [ad_conn user_id]
    set package_id [ad_conn package_id]

    set return_url "object-view?[export_url_vars object_id category_id]"

    set result {}
    set write_p [km_conn write_p]
    set admin_p [km_conn admin_p]
    set publish_p [km_conn publish_p]
    set public_p [km_conn public_p]
    set archived_p [km_conn archived_p]
    set archive_p [km_static object_type_archive_p [km_conn object_type_id]]
    set approval_p [km_static approval_p $package_id]
    set in_review_p [km_conn in_review_p]
    set review_state [km_conn review_state]

    # admins or users with write permission (if object is not archived)
    # can make object public/private
    if {$admin_p || ($approval_p && $publish_p)} {
	if {!$public_p} {
	    if {$approval_p && $in_review_p} {
                if {[string equal $review_state library_review]} {
		    lappend result "<font color=red>\[Private\]</font> <font color=green>\[In review\]</font> <a href=\"object-publish?[export_url_vars object_id category_id return_url]\">Make public</a>"
		} else {
		    lappend result "<font color=red>\[Private\]</font> <font color=green>\[Needs clarification\]</font> <a href=\"object-publish?[export_url_vars object_id category_id return_url]\">Make public</a>"
		}
	    } else {
		lappend result "<font color=red>\[Private\]</font> <a href=\"object-publish?[export_url_vars object_id category_id return_url]\">Make public</a>"
	    }
	} else {
	    lappend result "\[Public\] <a href=\"object-unpublish?[export_url_vars object_id category_id return_url]&[csrf::link_token]\">Make private</a>"
	}
    } elseif {$write_p && !$archived_p} {
	if { !$public_p } {
            if {!$in_review_p && $approval_p} {
		lappend result "<font color=red>\[Private\]</font> <a href=\"object-propose-public?[export_url_vars object_id category_id return_url]\">Propose to publish</a>"
	    } elseif {!$in_review_p} {
		lappend result "<font color=red>\[Private\]</font> <a href=\"object-publish?[export_url_vars object_id category_id return_url]\">Make public</a>"
            } else {
                # The state of a workflow task is user-dependant, hence we need the
                # else clause. A task might be in review, but there's nothing going 
                # on for the user.
		db_1row get_workflow_task {
		    select distinct(ut.task_id)
		    from wf_cases c, wf_user_tasks ut
		    where c.object_id = :object_id
		    and c.case_id = ut.case_id
		    and c.state = 'active'
		    and rownum = 1
		    order by ut.task_id desc
		}
                if {[string equal $review_state library_review]} {
                    lappend result "<font color=red>\[Private\]</font> <a href=\"../approval-tasks/task?[export_url_vars task_id]\">In review</a>"
                } elseif {[string equal $review_state library_clarify]} {
                    lappend result "<font color=red>\[Private\]</font> Review: <a href=\"../approval-tasks/task?[export_url_vars task_id]\">Needs clarification</a>"
                } else {
                    lappend result "<font color=red>\[Private\]</font> <font color=green>\[In review\]</font>"
                }
            }
	} elseif {!$approval_p} {
	    lappend result "\[Public\] <a href=\"object-unpublish?[export_url_vars object_id category_id return_url]\">Make private</a>"
	} else {
	    lappend result "\[Public\]"
	}
    } else {
	if { $in_review_p } {
	    lappend result "<font color=red>\[Private\]</font> <font color=green>\[In review\]</font>"
	} elseif { !$public_p } {
	    lappend result "<font color=red>\[Private\]</font>"
	}
    }
    ## users with write permission can archive/unarchive object
    ## (and admins and publishers)
    if {$write_p || $admin_p || ($approval_p && $publish_p)} {
	if {$archived_p} {
	    lappend result "<font color=red>\[Archived\]</font> <a href=\"object-unarchive?[export_url_vars object_id category_id return_url]\">Remove from archive</a>"
	}
	if {$archive_p && !$archived_p && ($admin_p || $publish_p || !$approval_p || (!$public_p && !$in_review_p))} {
	    lappend result "<a href=\"object-archive?[export_url_vars object_id category_id return_url]\">Put in archive</a>"
	}
    }
    # admins or users with write permission (if object is not archived)
    # can change permissions on the object
    if {$admin_p || ($write_p && !$archived_p)} {
	lappend result "<a href=\"object-access?[export_url_vars object_id category_id]\">Access Control</a>"
    }
    # users with delete permissions (owner or admins) can delete an object
    if [km_conn delete_p] {
	if {[empty_string_p $category_id] || $category_id == "none"} {
	    set return_delete "browse-one-type?object_type_id=[km_conn object_type_id]"
	} else {
	    set return_delete "browse-one-category?object_type_id=[km_conn object_type_id]&category_id=$category_id"
	}
	lappend result "<a href=\"object-delete?[export_url_vars object_id category_id return_delete]\">Delete Object</a>"
    }

    # admins or publishers have access to the approval task
    if {$approval_p && $in_review_p && ($publish_p || $admin_p)} {
	db_1row get_workflow_task {
	    select distinct(ut.task_id)
	    from wf_cases c, wf_user_tasks ut
	    where c.object_id = :object_id
	    and c.case_id = ut.case_id
	    and c.state = 'active'
	    and rownum = 1
	    order by ut.task_id desc
	}
	lappend result "<a href=\"../approval-tasks/task?[export_url_vars task_id]\">Manage Approval</a>"
    }

    return [join $result " | "]
}

ad_proc km_shorten_question { question } {
    Shortens a question up to the question mark in the question.
} {
    set line_end [string first ? $question]
    if { $line_end >= 0 }  {
	return [string range $question 0 $line_end]
    } else {
	return $question
    }
}

#--------------------------------------------------------


ad_proc util_trim_string_with_hrefs { string_to_trim { final_length 200 } } {
    Will return the trimmed string without cutting off an href before the </a> tag.
} {

    if {[string length $string_to_trim] < $final_length} {
	return $string_to_trim
    }
    
    set string_trimmed [string range $string_to_trim 0 $final_length]
    set string_trimmed_lower [string tolower $string_trimmed]

    # find the last occurance of "<a href"
    
    set last_href [string last "<a href" $string_trimmed_lower]

    if { $last_href != -1 } {

	# find the last occurance of </a>
	set last_href_close [string last "</a>" $string_trimmed_lower]

	# if there is no </a> after the last <a href then cut off 
	# the string at the last <a href

	if {$last_href > $last_href_close} {
	    set string_trimmed [string range $string_trimmed 0 [expr $last_href - 1]]
	}
    }

    return "$string_trimmed..."
}


ad_proc util_make_href_and_mailto_links { text } {
    Changes http:// and https:// to links and email formats to mailto's
} {
    # (bd) if there are any A HREF or IMG SRCs this guy know how to use them
    # so we don't touch his HTML. It may be hard to see but I am using
    # carriage returns, # newlines, tabs and spaces (in that order) in these
    # regexps:
    set text " $text"
    if {!([regexp -nocase {<[
	 ]*a[
	 ]*href[
	 ]*=} $text] || [regexp -nocase {<[
	 ]*img[
	 ]*src[
	 ]*=} $text])} {
       # if something is " http://" or " https://"
       # we assume it is a link to an outside source. 

       # (bd) The only purpose of thiese sTaRtUrL and
       # eNdUrL markers is to get rid of trailing dots,
       # commas and things like that.  Note that there
       # is a TAB before and after each marker.

       regsub -nocase -all {&quot;} $text {"} text
       regsub -nocase -all {&#34;} $text {"} text
       regsub -nocase -all {([^a-zA-Z0-9]+)(http://[^\(\)"<>
	 ]+)} $text {\1	sTaRtUrL\2eNdUrL	} text
       regsub -nocase -all {([^a-zA-Z0-9]+)(https://[^\(\)"<>
	 ]+)} $text {\1	sTaRtUrL\2eNdUrL	} text
       regsub -nocase -all {([^a-zA-Z0-9]+)(ftp://[^\(\)"<>
	 ]+)} $text {\1	sTaRtUrL\2eNdUrL	} text
   
       # email links have the form xxx@xxx.xxx
       regsub -nocase -all {([^a-zA-Z0-9]+)([^\(\)
	 :;,@<>]+@[^\(\)
	 .:;,@<>]+[.][^\(\)
	 :;,@<>]+)} $text {\1	sTaRtEmAiL\2eNdEmAiL	} text
       #
       regsub -all {([]!?.:;,<>\(\)\}-]+)(eNdUrL	)} $text {\2\1} text
       regsub -all {([]!?.:;,<>\(\)\}-]+)(eNdEmAiL	)} $text {\2\1} text
       regsub -all {	sTaRtUrL([^	]*)eNdUrL	} $text {<a href="\1">\1</a>} text
       regsub -all {	sTaRtEmAiL([^	]*)eNdEmAiL	} $text {<a href="mailto:\1">\1</a>} text
    }
    return [string trimleft $text]
}

ad_proc util_make_href_and_mailto_links_and_convert_to_html { text {html_p 0} } {

    util_make_href_and_mailto_links interferes badly with util_convert_to_html.
    That's why I had to write a special procedure that does both things correctly.
    See those two procs for details.

} {

    if { $html_p } { return $text }

    set text " $text"
    # if something is " http://" or " https://"
    # we assume it is a link to an outside source. 

    # (bd) The only purpose of thiese sTaRtUrL and
    # eNdUrL markers is to get rid of trailing dots,
    # commas and things like that.  Note that there
    # is a TAB before and after each marker.

    regsub -nocase -all {([^a-zA-Z0-9]+)(http://[^\(\)"<>
    ]+)} $text {\1	sTaRtUrL\2eNdUrL	} text
    regsub -nocase -all {([^a-zA-Z0-9]+)(https://[^\(\)"<>
    ]+)} $text {\1	sTaRtUrL\2eNdUrL	} text
    regsub -nocase -all {([^a-zA-Z0-9]+)(ftp://[^\(\)"<>
    ]+)} $text {\1	sTaRtUrL\2eNdUrL	} text
   
    # email links have the form xxx@xxx.xxx
    regsub -nocase -all {([^a-zA-Z0-9]+)([^\(\)
    :;,@<>]+@[^\(\)
    .:;,@<>]+[.][^\(\)
    :;,@<>]+)} $text {\1	sTaRtEmAiL\2eNdEmAiL	} text
    # At this point, before inserting some of our own <, >, and "'s
    # we quote the ones entered by the user:
    set text [util_quotehtml $text]
    # turn CRLFCRLF into <P>
    if { [regsub -all "\015\012\015\012" $text "<p>" text] == 0 } {
	# try LFLF
	if { [regsub -all "\012\012" $text "<p><p>" text] == 0 } {
	    # try CRCR
	    regsub -all "\015\015" $text "<p><p>" text
	}
    }
    # turn CRLF into <BR>
    if { [regsub -all "\015\012" $text "<br>" text] == 0 } {
	# try LF
	if { [regsub -all "\012" $text "<br>" text] == 0 } {
	    # try CR
	    regsub -all "\015" $text "<br>" text
	}
    }
    # Dress the links and emails with A HREF
    regsub -all {([]!?.:;,<>\(\)\}-]+)(eNdUrL	)} $text {\2\1} text
    regsub -all {([]!?.:;,<>\(\)\}-]+)(eNdEmAiL	)} $text {\2\1} text
    regsub -all {	sTaRtUrL([^	]*)eNdUrL	} $text {<a href="\1">\1</a>} text
    regsub -all {	sTaRtEmAiL([^	]*)eNdEmAiL	} $text {<a href="mailto:\1">\1</a>} text
    return [string trimleft $text]
}

ad_proc library_icon_height {} "Returns the height for an object icon" {
    return "50"
}

ad_proc library_icon_width {} "Returns the width for an object icon" {
    return "50"
}

ad_proc space {{-size 20}} "returns an html space" {

    set space ""
    set i 0
    while { $i < $size } {
        append space "&nbsp;"
        incr i
    }
    return $space

}
