# /tcl/km-psn-procs.tcl
#
# People Sharenet Procs.
#
# Created by timo@arsdigita.com in late October 2001
#
# $Id: km-psn-procs.tcl,v 1.1 2003/07/02 12:19:41 peterm Exp $

ad_proc -public portal_saved_applications { cf } {
    Displays the list of the saved and unfinished applications for demand
    descriptions.
} {
    set user_id [ad_conn user_id]
    set return_url "[ad_conn url]?[ad_conn query]"

    template::multirow create saved_applications project_name demand_name application_id library_url return_url

    db_foreach portal_saved_applications_10 {
	select o1.one_line_description as demand_name, a.object_id,
	       o2.one_line_description as project_name, a.application_id
	from sn_objects o1, sn_objects o2, psn_res_applications a
	where a.user_id = :user_id
	and a.sent_p = 'f'
	and o1.object_id = a.object_id
	and o2.object_id = a.resource_req_id
    } {
	db_1row portal_saved_applications_20 {
	    select site_node.url(sn.node_id) as library_url
	    from site_nodes sn, acs_objects_description dsc
	    where sn.object_id = dsc.package_id
	    and dsc.object_id = :object_id
	    and rownum = 1
	}
	template::multirow append saved_applications $project_name $demand_name $application_id $library_url $return_url
    }
}

ad_proc -public portal_approval_processes { cf } {
    Displays the list of the approval workflows.
} {
    set user_id [ad_conn user_id]

    set package_id [portal_arg $cf package_id]
    if [empty_string_p $package_id] {
	# grab the community_id
	set where_sql "o.context_id in (select package_id
                  from sn_community_mounts_all cm
                  where cm.community_id = :community_id
    	          and acs_permission.permission_p(package_id, :user_id, 'access') = 't')"
    } else {
	set where_sql "o.context_id = :package_id"
    }

    set last_package_id ""

    template::multirow create task_list task_id enabled_date enabled_time started_date started_time state object_id object_name object_type_id object_type_pretty workflow_key task_name task_url submitter_id submitter_name package_id package_name package_url

    db_transaction {
	if { [empty_string_p $package_id] } {
	    array set node [site_node_closest_ancestor_site_node "acs-subsite"] 
	    set community_id $node(package_id) 

	    db_dml check_package_access "
		insert into yp_package_access
		select p.package_id
		from sn_community_mounts cm, apm_packages p
		where cm.community_id = :community_id
		    and p.package_id = cm.package_id
		    and p.package_key = 'library'
		    and [tcl_permission_for_bind_vars_p "p.package_id" ":user_id" "'access'"] = 't'
	    "
	} else {
	    db_dml check_package_access_one {
		insert into yp_package_access
		select :package_id
		from dual
		where acs_permission.permission_p(:package_id, :user_id, 'access') = 't'
	    }
	}

	db_foreach get_workflows {
	    select t.task_id, t.state, c.object_id, o.object_type_id,
		t.enabled_date, t.started_date,
		to_char(t.enabled_date, 'HH24;MI') as enabled_time_pretty,
		to_char(t.started_date, 'HH24:MI') as started_time_pretty,
		ot.pretty_name as object_type_pretty, o.one_line_description as object_name,
		c.workflow_key, t.transition_name as task_name, u.user_id as submitter_id,
		u.first_names || ' ' || u.last_name as submitter_name,
		o.context_id as package_id
	    from wf_cases c, sn_objects o, sn_object_types ot, wf_user_tasks t,
		users u, acs_objects ao, yp_package_access ypa
	    where c.case_id = t.case_id
		and c.object_id = o.object_id
		and ot.object_type_id = o.object_type_id
		and t.user_id = :user_id
		and (t.state = 'enabled' or (t.state = 'started' and t.holding_user = t.user_id))
		and c.workflow_key = 'library_approval_wf'
		and ao.object_id = c.case_id
		and ao.creation_user = u.user_id
		and o.context_id = ypa.package_id
	    order by o.context_id, t.transition_name, t.enabled_date desc, ot.pretty_name, o.one_line_description
	} {
	    if {$package_id != $last_package_id} {
		set last_package_id $package_id
		db_1row get_package_data {
		    select site_node.url(n.node_id) as package_url, p.instance_name as package_name
		    from site_nodes n, apm_packages p
		    where n.object_id = :package_id
			and p.package_id = :package_id
			and rownum = 1
		}
	    }

	    template::multirow append task_list $task_id \
		[util_AnsiDatetoPrettyDate $enabled_date] $enabled_time_pretty \
		[util_AnsiDatetoPrettyDate $started_date] $started_time_pretty \
		$state $object_id $object_name $object_type_id $object_type_pretty \
		$workflow_key $task_name "../approval-tasks/task?task_id=$task_id" \
		$submitter_id $submitter_name $package_id $package_name $package_url
	}
    }
}

ad_proc -public km_sweeper_daily { } {
    db_transaction {
	km_sweeper_warn "outdated"
	km_sweeper_warn "expired"
	km_sweeper "outdated"
	km_sweeper "expired"
    }
}

ad_proc -private km_sweeper_warn { time } {
} {
    if {[string equal $time "expired"]} {
	set time_sql "o.expired_warning_date is null and o.public_until is not null and o.public_until + 1 - ot.sweeper_warning_time < sysdate"
    } else {
	set time_sql "o.outdated_warning_date is null and o.last_modified + ot.sweeper_outdated_time - ot.sweeper_warning_time < sysdate"
    }

    db_transaction {
	################
	## process the single objects
	################
	set count 0
	db_foreach get_outdated_objects_for_warning "
	select o.object_id, o.object_type_id, o.one_line_description as object_name,
	       ot.pretty_name as object_type, o.original_author_id, o.context_id,
	       ot.sweeper_action, (sysdate - o.last_modified) as days_unchanged,
	       ot.sweeper_warning_time
	from sn_objects o, sn_object_types ot
	where ot.object_type_id = o.object_type_id
	and o.public_p = 't'
	and o.archived_p = 'f'
	and o.expiration_date > sysdate
        and ot.sweeper = :time
	and ot.sweeper_warning_time > 0
	and $time_sql
	" {
	    lappend o_object_id $object_id
	    lappend o_object_type_id $object_type_id
	    lappend o_object_name $object_name
	    lappend o_object_type $object_type
	    lappend o_original_author_id $original_author_id
	    lappend o_context_id $context_id
	    lappend o_sweeper_action $sweeper_action
	    lappend o_days_unchanged $days_unchanged
	    lappend o_sweeper_warning_time $sweeper_warning_time
	    incr count
	}

	if {$count>0} {
	    foreach object_id $o_object_id object_type_id $o_object_type_id \
		    object_name $o_object_name object_type $o_object_type \
		    original_author_id $o_original_author_id \
		    context_id $o_context_id sweeper_action $o_sweeper_action \
		    days_unchanged $o_days_unchanged \
		    sweeper_warning_time $o_sweeper_warning_time {

		if [empty_string_p $object_name] {
		    set object_name "Unnamed $object_type"
		}

		##	    set mail_body [km_sweeper_single_objects_warning_mail $time $sweeper_action $object_type $object_name $object_id $days_unchanged $sweeper_warning_time]
		set mail_body "$object_name ([ad_url]/o/$object_id)\n"

		db_dml warn_single_object_outdated {
		    insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
		    values (:original_author_id, :context_id, :object_type_id, :object_id, :object_name, :mail_body)
		}
	    }

	    ## now register that warnings got send for these objects
	    db_dml object_warnings "
	        update sn_objects
	        set $time\_warning_date = trunc(sysdate)
	        where object_id in (select o.object_id
	                            from sn_objects o, sn_object_types ot
	                            where ot.object_type_id = o.object_type_id
	                            and o.public_p = 't'
	                            and o.archived_p = 'f'
	                            and ot.sweeper = :time
	                            and ot.sweeper_warning_time > 0
	                            and $time_sql )
	    "

	    unset o_object_id
	    unset o_object_type_id
	    unset o_object_name
	    unset o_object_type
	    unset o_original_author_id
	    unset o_context_id
	    unset o_sweeper_action
	    unset o_days_unchanged
	    unset o_sweeper_warning_time
	    
	}

	################
	## now send all the email, one per user, ordered by object_type
	################
	set mail ""
	set last_user_id ""
	set last_user_email ""
	set last_community_id ""
	set last_package_id ""
	set last_object_type_id ""
	set count 0
	db_foreach get_mail_bodies_per_user {
	    select u.user_id, u.email as user_email, s.content as mail_body,
	           ot.object_type_id, ot.pretty_plural as object_type,
	           p.instance_name as package_name, s.package_id,
	           p2.instance_name as community_name, ci.community_id,
	           ot.sweeper_warning_time, ot.sweeper_outdated_time
	    from km_sweeper s, users u, sn_object_types ot,
	         sn_community_instances ci, apm_packages p, apm_packages p2
	    where u.user_id = s.user_id
	    and s.object_type_id = ot.object_type_id
	    and ci.package_id = s.package_id
	    and p.package_id = s.package_id
	    and p2.package_id = ci.community_id
	    order by s.user_id, p2.instance_name, p.instance_name,
	             ot.pretty_plural, s.object_name
	} {
	    lappend o_user_id $user_id
	    lappend o_user_email $user_email
	    lappend o_mail_body $mail_body
	    lappend o_object_type_id $object_type_id
	    lappend o_object_type $object_type
	    lappend o_package_name $package_name
	    lappend o_package_id $package_id
	    lappend o_community_name $community_name
	    lappend o_community_id $community_id
	    lappend o_sweeper_warning_time $sweeper_warning_time
	    lappend o_sweeper_outdated_time [expr $sweeper_outdated_time - $sweeper_warning_time]
	    incr count
	}

	if {$count>0} {
	    foreach user_id $o_user_id user_email $o_user_email \
		    mail_body $o_mail_body object_type_id $o_object_type_id \
		    object_type $o_object_type package_name $o_package_name \
		    package_id $o_package_id community_name $o_community_name \
		    community_id $o_community_id \
		    sweeper_warning_time $o_sweeper_warning_time \
		    sweeper_outdated_time $o_sweeper_outdated_time {

		if {($last_user_id != $user_id) && ![empty_string_p $last_user_id]} {
		    set replacement_list \
			    [list \
			    type           $time \
			    mail_body      $mail \
			    recipient      $last_user_email]

		    et_queue_email [et_process "km_sweeper_warn" $replacement_list]
		    set last_user_id ""
		    set last_user_email ""
		    set last_community_id ""
		    set last_package_id ""
		    set last_object_type_id ""
		    set mail ""
		}
		if {$last_community_id != $community_id} {
		    append mail "Community: $community_name\n===========\n"
		    set last_community_id $community_id
		}
		if {$last_package_id != $package_id} {
		    append mail "\n$package_name:\n===========\n"
		    set last_package_id $package_id
		}
		if {$last_object_type_id != $object_type_id} {
		    append mail "\n$object_type:\n===========\n\n"
		    if {![string equal $time "expired"]} {
			append mail "The following objects haven't been changed in the last $sweeper_outdated_time days. "
		    }
		    append mail "Please change the objects in the next $sweeper_warning_time days:\n\n"
		    set last_object_type_id $object_type_id
		}
		append mail $mail_body
		set last_user_id $user_id
		set last_user_email $user_email
	    }
	    if ![empty_string_p $mail] {
		set replacement_list \
			[list \
			type           $time \
			mail_body      $mail \
			recipient      $user_email]

		et_queue_email [et_process "km_sweeper_warn" $replacement_list]
	    }

	    unset o_user_id
	    unset o_user_email
	    unset o_mail_body
	    unset o_object_type_id
	    unset o_object_type
	    unset o_package_name
	    unset o_package_id
	    unset o_community_name
	    unset o_community_id
	    unset o_sweeper_warning_time
	    unset o_sweeper_outdated_time
	}

	################
	## now send the email to object publisher
	## one per user, ordered by object_type
	################
	set mail ""
	set last_publisher_id ""
	set last_publisher_email ""
	set last_community_id ""
	set last_package_id ""
	set last_object_type_id ""
	set count 0
	db_foreach get_publisher_mail_bodies_per_user {
	    select u.user_id as publisher_id, u.email as publisher_email,
	           s.content as mail_body, ot.object_type_id,
	           ot.pretty_plural as object_type,
	           p.instance_name as package_name, s.package_id,
	           p2.instance_name as community_name, ci.community_id,
	           ot.sweeper_warning_time, ot.sweeper_outdated_time
	    from km_sweeper s, users u, sn_object_types ot, sn_objects o,
	         sn_community_instances ci, apm_packages p, apm_packages p2
	    where u.user_id = o.publisher_id
	    and o.publisher_id <> s.user_id
	    and s.object_id = o.object_id
	    and s.object_type_id = ot.object_type_id
	    and ci.package_id = s.package_id
	    and p.package_id = s.package_id
	    and p2.package_id = ci.community_id
	    and s.package_id in (select distinct ac.package_id
	                         from approval_coordinators ac)
	    order by s.user_id, p2.instance_name, p.instance_name,
	             ot.pretty_plural, s.object_name
	} {
	    lappend o_publisher_id $publisher_id
	    lappend o_publisher_email $publisher_email
	    lappend o_mail_body $mail_body
	    lappend o_object_type_id $object_type_id
	    lappend o_object_type $object_type
	    lappend o_package_name $package_name
	    lappend o_package_id $package_id
	    lappend o_community_name $community_name
	    lappend o_community_id $community_id
	    lappend o_sweeper_warning_time $sweeper_warning_time
	    lappend o_sweeper_outdated_time [expr $sweeper_outdated_time - $sweeper_warning_time]
	    incr count
	}

	if {$count>0} {
	    foreach publisher_id $o_publisher_id publisher_email $o_publisher_email \
		    mail_body $o_mail_body object_type_id $o_object_type_id \
		    object_type $o_object_type package_name $o_package_name \
		    package_id $o_package_id community_name $o_community_name \
		    community_id $o_community_id \
		    sweeper_warning_time $o_sweeper_warning_time \
		    sweeper_outdated_time $o_sweeper_outdated_time {

		if {($last_publisher_id != $publisher_id) && ![empty_string_p $last_publisher_id]} {
		    set replacement_list \
			    [list \
			    type           $time \
			    mail_body      $mail \
			    recipient      $last_publisher_email]

		    et_queue_email [et_process "km_sweeper_warn_publisher" $replacement_list]
		    set last_publisher_id ""
		    set last_publisher_email ""
		    set last_community_id ""
		    set last_package_id ""
		    set last_object_type_id ""
		    set mail ""
		}
		if {$last_community_id != $community_id} {
		    append mail "Community: $community_name\n===========\n"
		    set last_community_id $community_id
		}
		if {$last_package_id != $package_id} {
		    append mail "\n$package_name:\n===========\n"
		    set last_package_id $package_id
		}
		if {$last_object_type_id != $object_type_id} {
		    append mail "\n$object_type:\n===========\n\n"
		    if {![string equal $time "expired"]} {
			append mail "The following objects haven't been changed in the last $sweeper_outdated_time days. "
		    }
		    append mail "Please change the objects in the next $sweeper_warning_time days:\n\n"
		    set last_object_type_id $object_type_id
		}
		append mail $mail_body
		set last_publisher_id $publisher_id
		set last_publisher_email $publisher_email
	    }
	    if ![empty_string_p $mail] {
		set replacement_list \
			[list \
			type           $time \
			mail_body      $mail \
			recipient      $publisher_email]

		et_queue_email [et_process "km_sweeper_warn_publisher" $replacement_list]
	    }

	    unset o_publisher_id
	    unset o_publisher_email
	    unset o_mail_body
	    unset o_object_type_id
	    unset o_object_type
	    unset o_package_name
	    unset o_package_id
	    unset o_community_name
	    unset o_community_id
	    unset o_sweeper_warning_time
	    unset o_sweeper_outdated_time
	}

	###############
	## now cleanup: delete temp table
	###############
	db_dml km_sweeper_cleanup {
	    delete from km_sweeper
	}
    }
}

# time: expired / outdated
ad_proc -private km_sweeper { time } {
    Sets objects with outdated publication date to private and sends an email
    notification to the object owner
} {
    if {[string equal $time "expired"]} {
	set time_sql "(o.public_until is not null and o.public_until+1 < sysdate and (o.expired_warning_date is null or o.expired_warning_date + ot.sweeper_warning_time < sysdate))"
	set time_sql2 "(o2.public_until is not null and o2.public_until+1 < sysdate and (o2.expired_warning_date is null or o2.expired_warning_date + ot2.sweeper_warning_time < sysdate))"
    } else {
	set time_sql "(o.last_modified + ot.sweeper_outdated_time < sysdate and (o.outdated_warning_date is null or o.outdated_warning_date + ot.sweeper_warning_time < sysdate))"
	set time_sql2 "(o2.last_modified + ot2.sweeper_outdated_time < sysdate and (o2.outdated_warning_date is null or o2.outdated_warning_date + ot2.sweeper_warning_time < sysdate))"
    }

    db_transaction {
    #################
    ## deal with uncle objects that won't have public nephews any longer
    #################
    ## IMPORTANT NOTE: this supports only ONE hierarchy level in the sense
    ## that only the DIRECT uncle of outdated objects will get private
    ## - but then all it's descendants (multiple levels) will get private also
    #################
    set outdated_uncles [db_list get_uncles_need_private_status "
	select foh.parent
	from km_flat_object_hierarchy foh, sn_objects o, sn_object_types ot
	where foh.parent = o.object_id
	and o.public_p = 't'
	and o.archived_p = 'f'
	and o.expiration_date > sysdate
	and o.object_type_id = ot.object_type_id
	and not ((ot.sweeper = 'expired' and o.public_until is not null and o.public_until+1 < sysdate and (o.expired_warning_date is null or o.expired_warning_date + ot.sweeper_warning_time < sysdate))
	or (ot.sweeper = 'outdated' and o.last_modified + ot.sweeper_outdated_time < sysdate and (o.outdated_warning_date is null or o.outdated_warning_date + ot.sweeper_warning_time < sysdate)))
	and foh.link_type = 'uncle_nephew'
	and exists (select 1
	            from sn_questions q, sn_question_object_type_map qotm,
	                 sn_object_types ot3
	            where qotm.object_type_id = o.object_type_id
	            and qotm.question_id = q.question_id
	            and q.abstract_data_type = 'nephew_object'
	            and qotm.mandatory_p = 't'
	            and qotm.question_state = 'active'
	            and q.target_object_type_id = ot3.object_type_id
	            and ot3.sweeper = :time
	            and not exists (select 1
	                            from sn_question_link_map qlm, sn_links l,
                                         sn_objects o2, sn_object_types ot2
	                            where qlm.question_id = q.question_id
	                            and qlm.link_id = l.link_id
	                            and l.object_id_a = o.object_id
	                            and l.object_id_b = o2.object_id
	                            and o2.object_type_id = ot2.object_type_id
	                            and o2.public_p = 't'
	                            and o2.archived_p = 'f'
	                            and ot2.sweeper = :time
	                            and not $time_sql2 ) )
	order by foh.distance desc
    "]

    foreach outdated_uncle $outdated_uncles {
	# check if uncle is still public and get uncle data
	# get the questions that won't have any public nephews any longer
	# get all nephews that result in private uncle
	# get all other descendants
	# send email to uncle owner
	# loop over nephews again, but with different owner
	#   if different owner -> send email
	# loop over other descendants again, but with different owner
	#   if different owner -> send email
	# unpublish uncle and all descendants

	db_1row get_outdated_uncle_object_data {
	   select o.object_id as uncle_id, o.object_type_id as uncle_object_type_id,
	          o.one_line_description as uncle_name, ot.sweeper_action,
	          ot.pretty_name as uncle_object_type, o.public_p, o.archived_p,
	          u.email as uncle_author_email, u.user_id as uncle_author_id,
	          u.first_names || ' ' || u.last_name as uncle_author, o.context_id,
	          ci.community_id, o.publisher_id as uncle_publisher_id
	   from sn_objects o, sn_object_types ot, users u,
	        sn_community_instances ci
	   where ot.object_type_id = o.object_type_id
	   and u.user_id = o.original_author_id
	   and o.object_id = :outdated_uncle
	   and ci.package_id = o.context_id
       }

       if [empty_string_p $uncle_name] {
	   set uncle_name "Unnamed $uncle_object_type"
       }

       ## if uncle of this object already got private, so did this object, so check
       ## if it's still public
       if {$public_p == "t" && $archived_p == "f"} {

	   if {[string equal $time "expired"]} {
	       set archive_sql "o.archived_p = 'f'"
	       set archive_sql2 "o2.archived_p = 'f'"
	   } else {
	       set archive_sql "o.public_p = 't'"
	       set archive_sql2 "o2.public_p = 't'"
	   }

	   ## get all violated questions with outdated nephews
	   set question_ids [db_list get_violated_questions "
	       select q.question_id
	       from sn_questions q, sn_question_object_type_map qotm,
	            sn_objects o, sn_object_types ot
	       where q.abstract_data_type = 'nephew_object'
	       and qotm.question_id = q.question_id
	       and qotm.mandatory_p = 't'
	       and qotm.question_state = 'active'
	       and qotm.object_type_id = o.object_type_id
	       and o.object_id = :uncle_id
	       and q.target_object_type_id = ot.object_type_id
	       and ot.sweeper = :time
	       and not exists (select 1
	                       from sn_question_link_map qlm, sn_links l,
	                            sn_objects o2, sn_object_types ot2
	                       where qlm.question_id = q.question_id
	                       and l.link_id = qlm.link_id
	                       and l.object_id_a = :uncle_id
	                       and l.object_id_b = o2.object_id
	                       and o2.object_type_id = ot2.object_type_id
	                       and o2.public_p = 't'
	                       and o2.archived_p = 'f'
	                       and ot2.sweeper = :time
	                       and not $time_sql2 )
	   "]

	   ## get all outdated nephews for the violated questions
	   set last_object_type_id ""
	   set nephew_list ""
	   set nephew_ids [list]

	   db_foreach get_outdated_violating_nephews "
	       select o.object_id, o.object_type_id,
	              o.one_line_description as object_name,
	              ot.pretty_plural as object_type_plural,
	              ot.pretty_name as object_type
	       from sn_objects o, sn_object_types ot, sn_question_link_map qlm,
	            sn_links l
	       where ot.object_type_id = o.object_type_id
	       and o.public_p = 't'
	       and o.archived_p = 'f'
	       and ot.sweeper = :time
	       and $time_sql
	       and l.object_id_a = :uncle_id
	       and l.object_id_b = o.object_id
	       and l.link_id = qlm.link_id
	       and qlm.question_id in ([join $question_ids ,])
	       order by o.object_type_id
	   " {
	       if [empty_string_p $object_name] {
		   set object_name "Unnamed $object_type"
	       }

	       if {$last_object_type_id != $object_type_id} {
		   append nephew_list "\n$object_type_plural: $object_name"
	       } else {
		   append nephew_list ", $object_name"
	       }
	       set last_object_type_id $object_type_id
	       lappend nephew_ids $object_id
	   }

	   ## get all other public descendants
	   set last_object_type_id ""
	   set object_list ""

	   db_foreach get_all_other_public_descendants "
	       select o.object_id, o.object_type_id,
	              o.one_line_description as object_name,
	              ot.pretty_plural as object_type_plural,
	              ot.pretty_name as object_type
	       from sn_objects o, sn_object_types ot, km_flat_object_hierarchy foh
	       where ot.object_type_id = o.object_type_id
	       and $archive_sql
	       and foh.parent = :uncle_id
	       and o.object_id = foh.child
	       and o.object_id not in ([join $nephew_ids ,])
	       order by o.object_type_id
	   " {
	       if [empty_string_p $object_name] {
		   set object_name "Unnamed $object_type"
	       }

	       if {$last_object_type_id != $object_type_id} {
		   append object_list "\n$object_type_plural: $object_name"
	       } else {
		   append object_list ", $object_name"
	       }
	       set last_object_type_id $object_type_id
	   }

	   ## send email to object owner
	   set mail_body [km_sweeper_uncle_with_nephews_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $nephew_list $object_list]

	   db_dml uncle_with_nephews_outdated {
	       insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
	       values (:uncle_author_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
	   }

	   if {![empty_string_p $uncle_publisher_id] && $uncle_publisher_id != $uncle_author_id} {
	       db_dml uncle_with_nephews_outdated_publisher {
		   insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
		   values (:uncle_publisher_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
	       }
	   }

	   ## get all outdated nephews for the violated questions with different
	   ## owners and send them email
	   ##################
	   set last_author_id ""
	   set last_object_type_id ""
	   set outdated_nephew_list ""
	   set count 0

	   db_foreach get_outdated_violating_nephews_other_owner "
	       select o.object_id, o.object_type_id, o.original_author_id as author_id,
	              o.one_line_description as object_name, u.email as author_email,
	              ot.pretty_plural as object_type_plural,
                      ot.pretty_name as object_type
	       from sn_objects o, sn_object_types ot, users u
	       where ot.object_type_id = o.object_type_id
	       and u.user_id = o.original_author_id
	       and u.user_id <> :uncle_author_id
	       and o.object_id in ([join $nephew_ids ,])
	       order by u.user_id, o.object_type_id
	   " {
	       lappend o_object_id $object_id
	       lappend o_object_type_id $object_type_id
	       lappend o_author_id $author_id
	       lappend o_object_name $object_name
	       lappend o_author_email $author_email
	       lappend o_object_type_plural $object_type_plural
	       lappend o_object_type $object_type
	       incr count
	   }

	   if {$count>0} {
	       foreach object_id $o_object_id object_type_id $o_object_type_id \
		       author_id $o_author_id author_email $o_author_email \
		       object_name $o_object_name object_type $o_object_type \
		       object_type_plural $o_object_type_plural {

		   if [empty_string_p $object_name] {
		       set object_name "Unnamed $object_type"
		   }

		   if {($last_author_id != $author_id) && ![empty_string_p $last_author_id]} {
		       set mail_body [km_sweeper_nephew_objects_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $uncle_author $outdated_nephew_list $object_list]

		       db_dml uncle_with_nephews_outdated {
			   insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
			   values (:author_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
		       }

		       set last_object_type_id ""
		       set outdated_nephew_list ""
		   }
		   if {$last_object_type_id != $object_type_id} {
		       append outdated_nephew_list "\n\n$object_type_plural:"
		       set last_object_type_id $object_type_id
		       set last_author_id $author_id
		   }
		   append outdated_nephew_list "\n$object_name ([ad_url]/o/$object_id)"
	       }
	       if ![empty_string_p $outdated_nephew_list] {

		   set mail_body [km_sweeper_nephew_objects_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $uncle_author $outdated_nephew_list $object_list]

		   db_dml uncle_with_nephews_outdated {
		       insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
		       values (:author_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
		   }
	       }

	       unset o_object_id
	       unset o_object_type_id
	       unset o_author_id
	       unset o_author_email
	       unset o_object_name
	       unset o_object_type
	       unset o_object_type_plural
	   }

	   ## get all other descendants with different owners and send them email
	   set last_author_id ""
	   set last_object_type_id ""
	   set descendants_list ""
	   set count 0

	   db_foreach get_all_other_public_descendants_other_owner "
	       select o.object_id, o.object_type_id, o.original_author_id as author_id,
	              o.one_line_description as object_name, u.email as author_email,
	              ot.pretty_plural as object_type_plural,
                      ot.pretty_name as object_type
	       from sn_objects o, sn_object_types ot, km_flat_object_hierarchy foh,
	            users u
	       where ot.object_type_id = o.object_type_id
	       and u.user_id = o.original_author_id
	       and u.user_id <> :uncle_author_id
	       and u.user_id not in (select distinct o2.original_author_id
	                             from sn_objects o2
	                             where o2.object_id in ([join $nephew_ids ,]))
	       and $archive_sql
	       and foh.parent = :uncle_id
	       and o.object_id = foh.child
	       and o.object_id not in ([join $nephew_ids ,])
	       order by u.user_id, o.object_type_id
	   " {
	       lappend o_object_id $object_id
	       lappend o_object_type_id $object_type_id
	       lappend o_author_id $author_id
	       lappend o_object_name $object_name
	       lappend o_author_email $author_email
	       lappend o_object_type_plural $object_type_plural
	       lappend o_object_type $object_type
	       incr count
	   }

	   if {$count>0} {
	       foreach object_id $o_object_id object_type_id $o_object_type_id \
		       author_id $o_author_id author_email $o_author_email \
		       object_name $o_object_name object_type $o_object_type \
		       object_type_plural $o_object_type_plural {

		   if [empty_string_p $object_name] {
		       set object_name "Unnamed $object_type"
		   }

		   if {($last_author_id != $author_id) && ![empty_string_p $last_author_id]} {
		       set mail_body [km_uncles_descendant_objects_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $uncle_author $nephew_list $descendants_list]

		       db_dml uncle_with_nephews_outdated {
			   insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
			   values (:author_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
		       }
		       set last_object_type_id ""
		       set descendants_list ""
		   }
		   if {$last_object_type_id != $object_type_id} {
		       append descendants_list "\n\n$object_type_plural:"
		       set last_object_type_id $object_type_id
		       set last_author_id $author_id
		   }
		   append descendants_list "\n$object_name ([ad_url]/o/$object_id)"
	       }
	       if ![empty_string_p $descendants_list] {

		   set mail_body [km_uncles_descendant_objects_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $uncle_author $nephew_list $descendants_list]

		   db_dml uncle_with_nephews_outdated {
		       insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
		       values (:author_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
		   }
	       }

	       unset o_object_id
	       unset o_object_type_id
	       unset o_author_id
	       unset o_author_email
	       unset o_object_name
	       unset o_object_type
	       unset o_object_type_plural
	   }

	   ## get all outdated nephews for the violated questions with different
	   ## publishers and send them email
	   ##################
	   if {[empty_string_p $uncle_publisher_id]} {
	       set uncle_publisher_id 0
	   }
	   set last_publisher_id ""
	   set last_object_type_id ""
	   set outdated_nephew_list ""
	   set count 0

	   db_foreach get_outdated_violating_nephews_other_publisher "
	       select o.object_id, o.object_type_id, o.publisher_id,
	              o.one_line_description as object_name,
	              ot.pretty_plural as object_type_plural,
                      ot.pretty_name as object_type
	       from sn_objects o, sn_object_types ot
	       where ot.object_type_id = o.object_type_id
	       and o.publisher_id is not null
	       and o.publisher_id <> :uncle_publisher_id
	       and o.publisher_id <> :uncle_author_id
	       and o.object_id in ([join $nephew_ids ,])
	       and o.publisher_id not in (select distinct o2.original_author_id
	                                  from sn_objects o2, km_flat_object_hierarchy foh
	                                  where foh.child = o2.object_id
	                                  and foh.parent = :uncle_id
	                                  and $archive_sql2)
	       order by o.publisher_id, o.object_type_id
	   " {
	       lappend o_object_id $object_id
	       lappend o_object_type_id $object_type_id
	       lappend o_publisher_id $publisher_id
	       lappend o_object_name $object_name
	       lappend o_object_type_plural $object_type_plural
	       lappend o_object_type $object_type
	       incr count
	   }

	   if {$count>0} {
	       foreach object_id $o_object_id object_type_id $o_object_type_id \
		       publisher_id $o_publisher_id object_name $o_object_name \
		       object_type $o_object_type \
		       object_type_plural $o_object_type_plural {

		   if [empty_string_p $object_name] {
		       set object_name "Unnamed $object_type"
		   }

		   if {($last_publisher_id != $publisher_id) && ![empty_string_p $last_publisher_id]} {
		       set mail_body [km_sweeper_nephew_objects_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $uncle_author $outdated_nephew_list $object_list]

		       db_dml uncle_with_nephews_outdated_publisher {
			   insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
			   values (:publisher_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
		       }

		       set last_object_type_id ""
		       set outdated_nephew_list ""
		   }
		   if {$last_object_type_id != $object_type_id} {
		       append outdated_nephew_list "\n\n$object_type_plural:"
		       set last_object_type_id $object_type_id
		       set last_publisher_id $publisher_id
		   }
		   append outdated_nephew_list "\n$object_name ([ad_url]/o/$object_id)"
	       }
	       if ![empty_string_p $outdated_nephew_list] {

		   set mail_body [km_sweeper_nephew_objects_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $uncle_author $outdated_nephew_list $object_list]

		   db_dml uncle_with_nephews_outdated_publisher {
		       insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
		       values (:publisher_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
		   }
	       }

	       unset o_object_id
	       unset o_object_type_id
	       unset o_publisher_id
	       unset o_object_name
	       unset o_object_type
	       unset o_object_type_plural
	   }

	   ## get all other descendants with different publisher
	   ## and send them email
	   set last_publisher_id ""
	   set last_object_type_id ""
	   set descendants_list ""
	   set count 0

	   db_foreach get_all_other_public_descendants_other_publisher "
	       select o.object_id, o.object_type_id, o.publisher_id,
	              o.one_line_description as object_name,
	              ot.pretty_plural as object_type_plural,
                      ot.pretty_name as object_type
	       from sn_objects o, sn_object_types ot, km_flat_object_hierarchy foh
	       where ot.object_type_id = o.object_type_id
	       and o.publisher_id is not null
	       and o.publisher_id <> :uncle_author_id
	       and o.publisher_id not in (select distinct o2.publisher_id
	                                  from sn_objects o2
	                                  where o2.object_id in ([join $nephew_ids ,]))
	       and o.publisher_id not in (select distinct o2.original_author_id
	                                  from sn_objects o2, km_flat_object_hierarchy foh2
	                                  where foh2.child = o2.object_id
	                                  and foh2.parent = :uncle_id
	                                  and $archive_sql2)
	       and o.public_p = 't'
	       and foh.parent = :uncle_id
	       and o.object_id = foh.child
	       and o.object_id not in ([join $nephew_ids ,])
	       order by o.publisher_id, o.object_type_id
	   " {
	       lappend o_object_id $object_id
	       lappend o_object_type_id $object_type_id
	       lappend o_publisher_id $publisher_id
	       lappend o_object_name $object_name
	       lappend o_object_type_plural $object_type_plural
	       lappend o_object_type $object_type
	       incr count
	   }

	   if {$count>0} {
	       foreach object_id $o_object_id object_type_id $o_object_type_id \
		       publisher_id $o_publisher_id object_name $o_object_name \
		       object_type $o_object_type \
		       object_type_plural $o_object_type_plural {

		   if [empty_string_p $object_name] {
		       set object_name "Unnamed $object_type"
		   }

		   if {($last_publisher_id != $publisher_id) && ![empty_string_p $last_publisher_id]} {
		       set mail_body [km_uncles_descendant_objects_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $uncle_author $nephew_list $descendants_list]

		       db_dml uncle_with_nephews_outdated_publisher {
			   insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
			   values (:publisher_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
		       }
		       set last_object_type_id ""
		       set descendants_list ""
		   }
		   if {$last_object_type_id != $object_type_id} {
		       append descendants_list "\n\n$object_type_plural:"
		       set last_object_type_id $object_type_id
		       set last_publisher_id $publisher_id
		   }
		   append descendants_list "\n$object_name ([ad_url]/o/$object_id)"
	       }
	       if ![empty_string_p $descendants_list] {

		   set mail_body [km_uncles_descendant_objects_mail $community_id $time $sweeper_action $uncle_object_type $uncle_name $uncle_id $uncle_author $nephew_list $descendants_list]

		   db_dml uncle_with_nephews_outdated_publisher {
		       insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
		       values (:publisher_id, :context_id, :uncle_object_type_id, :uncle_id, :uncle_name, :mail_body)
		   }
	       }

	       unset o_object_id
	       unset o_object_type_id
	       unset o_publisher_id
	       unset o_object_name
	       unset o_object_type
	       unset o_object_type_plural
	   }

	   ## now finally unpublish / archive the uncle and all descendants
	   ##########
	   if {[string equal $sweeper_action "archive"]} {
	       foreach nephew_id $nephew_ids {
		   km_object_archive $nephew_id "object $time, archived by sweeper"
	       }
	       km_object_archive $uncle_id "linked objects $time, archived by sweeper"
	   } else {
	       km_object_unpublish $uncle_id
	   }
       }
   }

   #################
   ## deal with outdated parent objects
   #################
   set outdated_parents [db_list get_outdated_parent_objects "
        select foh.parent
	from km_flat_object_hierarchy foh, sn_objects o, sn_object_types ot
	where foh.parent = o.object_id
	and o.public_p = 't'
	and o.archived_p = 'f'
        and o.expiration_date > sysdate
        and ot.object_type_id = o.object_type_id
        and ot.sweeper = :time
	and $time_sql
	order by distance desc
   "]

   foreach outdated_parent $outdated_parents {
       # check if object is still public and get object data
       # loop over children, order by object_type
       #   append children to email
       # loop over children with different owner, order by owner, object type
       #   if different child owner -> send email
       # unpublish children, unpublish parent

       db_1row get_outdated_parent_object_data {
	   select o.object_id as parent_id, o.object_type_id as parent_object_type_id,
	          o.one_line_description as parent_name, ot.sweeper_action,
	          ot.pretty_name as parent_object_type, o.public_p, o.archived_p,
	          u.email as parent_author_email, u.user_id as parent_author_id,
	          u.first_names || ' ' || u.last_name as parent_author,
	          trunc(sysdate - o.last_modified) as days_unchanged,
	          o.context_id, ci.community_id,
	          o.publisher_id as parent_publisher_id
	   from sn_objects o, sn_object_types ot, users u,
	        sn_community_instances ci
	   where ot.object_type_id = o.object_type_id
	   and u.user_id = o.original_author_id
	   and o.object_id = :outdated_parent
	   and ci.package_id = o.context_id
       }

       if [empty_string_p $parent_name] {
	   set parent_name "Unnamed $parent_object_type"
       }

       ## if parent of this object already got private, so did this object, so check
       ## if it's still public
       if {$public_p == "t" && $archived_p == "f"} {

	   ## create list of all children to be made private for the email to
	   ## the owner of the parent
	   set last_object_type_id ""
	   set object_list ""

	   if {[string equal $sweeper_action "archive"]} {
	       ## we have to archive all children, as long as they can
	       ## be archived (=> cut the tree off at this point)
	       set sql "( select o2.object_id
	                 from km_flat_object_hierarchy foh, sn_objects o2, sn_object_types ot2
	                 where foh.parent = :parent_id
	                 and foh.child = o2.object_id
	                 and o2.archived_p = 'f'
	                 and o2.object_type_id = ot2.object_type_id
	                 and ot2.archive_p = 't' )
	               minus
	               ( select child as object_id
	                 from km_flat_object_hierarchy
	                 where parent in (select o2.object_id
	                     from km_flat_object_hierarchy foh, sn_objects o2,
	                          sn_object_types ot2
	                     where foh.parent = :parent_id
	                     and foh.child = o2.object_id
	                     and o2.archived_p = 'f'
	                     and o2.object_type_id = ot2.object_type_id
	                     and ot2.archive_p = 'f') )"

	       set sql "select o.object_id, o.object_type_id,
	              o.one_line_description as object_name,
	              ot.pretty_plural as object_type_plural,
	              ot.pretty_name as object_type, o.publisher_id
	       from sn_objects o, sn_object_types ot
	       where ot.object_type_id = o.object_type_id
	       and o.object_id in ($sql)
	       order by o.object_type_id"
	   } else {
	       ## we have to make all children private
	       set sql "select o.object_id, o.object_type_id,
	              o.one_line_description as object_name,
	              ot.pretty_plural as object_type_plural,
	              ot.pretty_name as object_type, o.publisher_id
	       from sn_objects o, sn_object_types ot, km_flat_object_hierarchy foh
	       where ot.object_type_id = o.object_type_id
	       and o.public_p = 't'
	       and foh.parent = :parent_id
	       and o.object_id = foh.child
	       order by o.object_type_id"
	   }

	   set child_list [list]
	   db_foreach get_all_children_of_parent $sql {
	       if [empty_string_p $object_name] {
		   set object_name "Unnamed $object_type"
	       }

	       if {$last_object_type_id != $object_type_id} {
		   append object_list "\n$object_type_plural: $object_name"
	       } else {
		   append object_list ", $object_name"
	       }
	       set last_object_type_id $object_type_id
	       lappend child_list $object_id
	   }


	   ## send email to object owner
	   if {![llength $child_list]} {
	       ## so there are no children that will be put in the archive
	       ## so treat the parent as a single object
	       set mail_body [km_sweeper_single_objects_mail $community_id $time $sweeper_action $parent_object_type $parent_name $parent_id $days_unchanged]

	       db_dml object_without_children_outdated {
		   insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
		   values (:parent_author_id, :context_id, :parent_object_type_id, :parent_id, :parent_name, :mail_body)
	       }

	       if {![empty_string_p $parent_publisher_id] && $parent_publisher_id != $parent_author_id} {
		   db_dml object_without_children_outdated {
		       insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
		       values (:parent_publisher_id, :context_id, :parent_object_type_id, :parent_id, :parent_name, :mail_body)
		   }
	       }

	   } else {
	       ## so there are actually children
	       set mail_body [km_sweeper_object_with_children_mail $community_id $time $sweeper_action $parent_object_type $parent_name $parent_id $days_unchanged $object_list]

	       db_dml object_with_children_outdated {
		   insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
		   values (:parent_author_id, :context_id, :parent_object_type_id, :parent_id, :parent_name, :mail_body)
	       }

	       if {![empty_string_p $parent_publisher_id] && $parent_publisher_id != $parent_author_id} {
		   db_dml object_with_children_outdated_publisher {
		       insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
		       values (:parent_publisher_id, :context_id, :parent_object_type_id, :parent_id, :parent_name, :mail_body)
		   }
	       }

	       set last_author_id ""
	       set last_object_type_id ""
	       set object_list ""
	       set count 0

	       ## if children owner is not the same as parent owner, sent him
	       ## an email
	       #############
	       db_foreach get_all_children_of_parent_other_owner "
	           select o.object_id, o.object_type_id, o.original_author_id as author_id,
	                  o.one_line_description as object_name,
	                  ot.pretty_plural as object_type_plural,
	                  ot.pretty_name as object_type
	           from sn_objects o, sn_object_types ot
	           where ot.object_type_id = o.object_type_id
	           and o.original_author_id <> :parent_author_id
	           and o.object_id in ([join $child_list ,])
	           order by o.original_author_id, o.object_type_id
	       " {
		   lappend o_object_id $object_id
		   lappend o_object_type_id $object_type_id
		   lappend o_author_id $author_id
		   lappend o_object_name $object_name
		   lappend o_object_type_plural $object_type_plural
		   lappend o_object_type $object_type
		   incr count
	       }

	       if {$count>0} {
		   foreach object_id $o_object_id object_type_id $o_object_type_id \
			   author_id $o_author_id object_name $o_object_name \
			   object_type $o_object_type \
			   object_type_plural $o_object_type_plural {

		       if [empty_string_p $object_name] {
			   set object_name "Unnamed $object_type"
		       }

		       if {($last_author_id != $author_id) && ![empty_string_p $last_author_id]} {
			   set mail_body [km_sweeper_child_object_mail $community_id $time $sweeper_action $parent_object_type $parent_name $parent_id $parent_author $days_unchanged $object_list]

			   db_dml child_objects_outdated {
			       insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
			       values (:author_id, :context_id, :parent_object_type_id, :parent_id, :parent_name, :mail_body)
			   }
			   set last_object_type_id ""
			   set object_list ""
		       }
		       if {$last_object_type_id != $object_type_id} {
			   append object_list "\n\n$object_type_plural:"
			   set last_object_type_id $object_type_id
			   set last_author_id $author_id
		       }
		       append object_list "\n$object_name ([ad_url]/o/$object_id)"
		   }
		   if ![empty_string_p $object_list] {
		       set mail_body [km_sweeper_child_object_mail $community_id $time $sweeper_action $parent_object_type $parent_name $parent_id $parent_author $days_unchanged $object_list]

		       db_dml child_objects_outdated {
			   insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
			   values (:author_id, :context_id, :parent_object_type_id, :parent_id, :parent_name, :mail_body)
		       }
		   }

		   unset o_object_id
		   unset o_object_type_id
		   unset o_author_id
		   unset o_object_name
		   unset o_object_type
		   unset o_object_type_plural
	       }

	       set last_publisher_id ""
	       set last_object_type_id ""
	       set object_list ""
	       set count 0

	       ## if children publisher is not the same as of parent, sent him
	       ## an email
	       #############
	       if {[empty_string_p $parent_publisher_id]} {
		   set parent_publisher_id 0
	       }
	       db_foreach get_all_children_of_parent_other_publisher "
	           select o.object_id, o.object_type_id,
	                  o.one_line_description as object_name,
	                  ot.pretty_plural as object_type_plural,
	                  ot.pretty_name as object_type, o.publisher_id
	           from sn_objects o, sn_object_types ot
	           where ot.object_type_id = o.object_type_id
	           and o.publisher_id <> :parent_publisher_id
	           and o.publisher_id <> :parent_author_id
	           and o.publisher_id is not null
	           and o.object_id in ([join $child_list ,])
	           and o.publisher_id not in (select distinct o2.original_author_id
	                                      from sn_objects o2
	                                      where o2.object_id in ([join $child_list ,]))
	           order by o.publisher_id, o.object_type_id
	       " {
		   lappend o_object_id $object_id
		   lappend o_object_type_id $object_type_id
		   lappend o_publisher_id $publisher_id
		   lappend o_object_name $object_name
		   lappend o_object_type_plural $object_type_plural
		   lappend o_object_type $object_type
		   incr count
	       }

	       if {$count>0} {
		   foreach object_id $o_object_id object_type_id $o_object_type_id \
			   publisher_id $o_publisher_id object_name $o_object_name \
			   object_type $o_object_type \
			   object_type_plural $o_object_type_plural {

		       if [empty_string_p $object_name] {
			   set object_name "Unnamed $object_type"
		       }

		       if {($last_publisher_id != $publisher_id) && ![empty_string_p $last_publisher_id]} {
			   set mail_body [km_sweeper_child_object_mail $community_id $time $sweeper_action $parent_object_type $parent_name $parent_id $parent_author $days_unchanged $object_list]

			   db_dml child_objects_outdated_publisher {
			       insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
			       values (:publisher_id, :context_id, :parent_object_type_id, :parent_id, :parent_name, :mail_body)
			   }
			   set last_object_type_id ""
			   set object_list ""
		       }
		       if {$last_object_type_id != $object_type_id} {
			   append object_list "\n\n$object_type_plural:"
			   set last_object_type_id $object_type_id
			   set last_publisher_id $publisher_id
		       }
		       append object_list "\n$object_name ([ad_url]/o/$object_id)"
		   }
		   if ![empty_string_p $object_list] {
		       set mail_body [km_sweeper_child_object_mail $community_id $time $sweeper_action $parent_object_type $parent_name $parent_id $parent_author $days_unchanged $object_list]

		       db_dml child_objects_outdated_publisher {
			   insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
			   values (:publisher_id, :context_id, :parent_object_type_id, :parent_id, :parent_name, :mail_body)
		       }
		   }

		   unset o_object_id
		   unset o_object_type_id
		   unset o_publisher_id
		   unset o_object_name
		   unset o_object_type
		   unset o_object_type_plural
	       }
	   }

	   ## now finally archive/publish the objects
	   if {[string equal $sweeper_action "archive"]} {
	       km_object_archive $parent_id "object $time, archived by sweeper"
	   } else {
	       km_object_unpublish $parent_id
	   }
       }
   }

    ################
    ## now process the single objects
    ################
    set count 0
    db_foreach get_outdated_single_objects "
	select o.object_id, o.object_type_id, o.one_line_description as object_name,
	       ot.pretty_name as object_type, o.original_author_id, o.context_id,
	       ot.sweeper_action, trunc(sysdate - o.last_modified) as days_unchanged,
               ci.community_id, o.publisher_id
	from sn_objects o, sn_object_types ot, sn_community_instances ci
	where ot.object_type_id = o.object_type_id
	and o.public_p = 't'
	and o.archived_p = 'f'
        and o.expiration_date > sysdate
        and ot.sweeper = :time
	and $time_sql
        and ci.package_id = o.context_id
    " {
	lappend o_object_id $object_id
	lappend o_object_type_id $object_type_id
	lappend o_object_name $object_name
	lappend o_object_type $object_type
	lappend o_original_author_id $original_author_id
	lappend o_context_id $context_id
	lappend o_sweeper_action $sweeper_action
	lappend o_days_unchanged $days_unchanged
	lappend o_community_id $community_id
	lappend o_publisher_id $publisher_id
	incr count
    }

    if {$count>0} {
	foreach object_id $o_object_id object_type_id $o_object_type_id \
		object_name $o_object_name object_type $o_object_type \
		original_author_id $o_original_author_id context_id $o_context_id \
		sweeper_action $o_sweeper_action days_unchanged $o_days_unchanged \
		community_id $o_community_id publisher_id $o_publisher_id {

	    if [empty_string_p $object_name] {
		set object_name "Unnamed $object_type"
	    }

	    set mail_body [km_sweeper_single_objects_mail $community_id $time $sweeper_action $object_type $object_name $object_id $days_unchanged]

	    db_dml single_object_outdated {
		insert into km_sweeper (user_id, package_id, object_type_id, object_id, object_name, content)
		values (:original_author_id, :context_id, :object_type_id, :object_id, :object_name, :mail_body)
	    }

	    if {![empty_string_p $publisher_id] && $publisher_id != $original_author_id} {
		db_dml single_object_outdated_publisher {
		    insert into km_sweeper_publisher (user_id, package_id, object_type_id, object_id, object_name, content)
		    values (:publisher_id, :context_id, :object_type_id, :object_id, :object_name, :mail_body)
		}
	    }

	    if {[string equal $sweeper_action "archive"]} {
		km_object_archive $object_id "object $time, archived by sweeper"
	    } else {
		km_object_unpublish $object_id
	    }
	}

	unset o_object_id
	unset o_object_type_id
	unset o_object_name
	unset o_object_type
	unset o_original_author_id
	unset o_context_id
	unset o_sweeper_action
	unset o_days_unchanged
	unset o_community_id
	unset o_publisher_id
    }

    ###########################
    ## now send all the email, one per user, ordered by object_type
    ###########################
    set mail ""
    set last_user_id ""
    set last_user_email ""
    set last_community_id ""
    set last_package_id ""
    set last_object_type_id ""
    set count 0
    db_foreach get_mail_bodies_per_user {
	select u.user_id, u.email as user_email, s.content as mail_body,
	       ot.object_type_id, ot.pretty_plural as object_type,
	       p.instance_name as package_name, s.package_id,
	       p2.instance_name as community_name, ci.community_id
	from km_sweeper s, users u, sn_object_types ot,
	     sn_community_instances ci, apm_packages p, apm_packages p2
	where u.user_id = s.user_id
	and s.object_type_id = ot.object_type_id
	and ci.package_id = s.package_id
	and p.package_id = s.package_id
	and p2.package_id = ci.community_id
	order by s.user_id, p2.instance_name, p.instance_name,
	      ot.pretty_plural, s.object_name
    } {
	lappend o_user_id $user_id
	lappend o_user_email $user_email
	lappend o_mail_body $mail_body
	lappend o_object_type_id $object_type_id
	lappend o_object_type $object_type
	lappend o_package_name $package_name
	lappend o_package_id $package_id
	lappend o_community_name $community_name
	lappend o_community_id $community_id
	incr count
    }

    if {$count>0} {
	foreach user_id $o_user_id user_email $o_user_email \
		mail_body $o_mail_body object_type_id $o_object_type_id \
		object_type $o_object_type package_name $o_package_name \
		package_id $o_package_id community_name $o_community_name \
		community_id $o_community_id {

	    if {($last_user_id != $user_id) && ![empty_string_p $last_user_id]} {
		set replacement_list \
			[list \
			type           $time \
			mail_body      $mail \
			recipient      $last_user_email]

		et_queue_email [et_process "km_sweeper" $replacement_list]
		set last_user_id ""
		set last_user_email ""
		set last_community_id ""
		set last_package_id ""
		set last_object_type_id ""
		set mail ""
	    }
	    if {$last_community_id != $community_id} {
		append mail "Community: $community_name\n===========\n\n"
		set last_community_id $community_id
	    }
	    if {$last_package_id != $package_id} {
		append mail "$package_name:\n===========\n\n"
		set last_package_id $package_id
	    }
	    if {$last_object_type_id != $object_type_id} {
		append mail "$object_type:\n===========\n\n"
		set last_object_type_id $object_type_id
	    }
	    append mail "$mail_body\n---------------\n\n"
	    set last_user_id $user_id
	    set last_user_email $user_email
	}
	if ![empty_string_p $mail] {
	    set replacement_list \
		    [list \
		    type           $time \
		    mail_body      $mail \
		    recipient      $user_email]

	    et_queue_email [et_process "km_sweeper" $replacement_list]
	}

	unset o_user_id
	unset o_user_email
	unset o_mail_body
	unset o_object_type_id
	unset o_object_type
	unset o_package_name
	unset o_package_id
	unset o_community_name
	unset o_community_id
    }

    ###########################
    ## now send the email to object publisher
    ## one per user, ordered by object_type
    ###########################
    set mail ""
    set last_publisher_id ""
    set last_publisher_email ""
    set last_community_id ""
    set last_package_id ""
    set last_object_type_id ""
    set count 0
    db_foreach get_mail_bodies_per_publisher {
	select u.user_id as publisher_id, u.email as publisher_email,
	       s.content as mail_body, ot.object_type_id,
	       ot.pretty_plural as object_type, p.instance_name as package_name,
	       s.package_id, p2.instance_name as community_name, ci.community_id
	from km_sweeper_publisher s, users u, sn_object_types ot,
	     sn_community_instances ci, apm_packages p, apm_packages p2
	where u.user_id = s.user_id
	and s.object_type_id = ot.object_type_id
	and ci.package_id = s.package_id
	and p.package_id = s.package_id
	and p2.package_id = ci.community_id
	order by u.user_id, p2.instance_name, p.instance_name,
	      ot.pretty_plural, s.object_name
    } {
	lappend o_publisher_id $publisher_id
	lappend o_publisher_email $publisher_email
	lappend o_mail_body $mail_body
	lappend o_object_type_id $object_type_id
	lappend o_object_type $object_type
	lappend o_package_name $package_name
	lappend o_package_id $package_id
	lappend o_community_name $community_name
	lappend o_community_id $community_id
	incr count
    }

    if {$count>0} {
	foreach publisher_id $o_publisher_id publisher_email $o_publisher_email \
		mail_body $o_mail_body object_type_id $o_object_type_id \
		object_type $o_object_type package_name $o_package_name \
		package_id $o_package_id community_name $o_community_name \
		community_id $o_community_id {

	    if {($last_publisher_id != $publisher_id) && ![empty_string_p $last_publisher_id]} {
		set replacement_list \
			[list \
			type           $time \
			mail_body      $mail \
			recipient      $last_publisher_email]

		et_queue_email [et_process "km_sweeper_publisher" $replacement_list]
		set last_publisher_id ""
		set last_publisher_email ""
		set last_community_id ""
		set last_package_id ""
		set last_object_type_id ""
		set mail ""
	    }
	    if {$last_community_id != $community_id} {
		append mail "Community: $community_name\n===========\n\n"
		set last_community_id $community_id
	    }
	    if {$last_package_id != $package_id} {
		append mail "$package_name:\n===========\n\n"
		set last_package_id $package_id
	    }
	    if {$last_object_type_id != $object_type_id} {
		append mail "$object_type:\n===========\n\n"
		set last_object_type_id $object_type_id
	    }
	    append mail "$mail_body\n---------------\n\n"
	    set last_publisher_id $publisher_id
	    set last_publisher_email $publisher_email
	}
	if ![empty_string_p $mail] {
	    set replacement_list \
		    [list \
		    type           $time \
		    mail_body      $mail \
		    recipient      $publisher_email]

	    et_queue_email [et_process "km_sweeper_publisher" $replacement_list]
	}

	unset o_publisher_id
	unset o_publisher_email
	unset o_mail_body
	unset o_object_type_id
	unset o_object_type
	unset o_package_name
	unset o_package_id
	unset o_community_name
	unset o_community_id
    }

    ##############
    ## now cleanup: delete temp tables
    ##############
    db_dml km_sweeper_cleanup {
	delete from km_sweeper
    }
    db_dml km_sweeper_publisher_cleanup {
	delete from km_sweeper_publisher
    }
## end of transaction
}
}


ad_proc km_sweeper_uncle_with_nephews_mail { community_id time action object_type object_name object_id nephew_list object_list } {
} {
    set template "km_sweeper_uncle_$action\_$time"
    set replace_list [list \
	    ad_url [ad_url] \
	    object_type $object_type \
	    object_name $object_name \
	    object_id $object_id \
	    object_list $object_list \
	    nephew_list $nephew_list]
    set result [et_process -community_id $community_id $template $replace_list]

    return [ns_set get $result body]
}


ad_proc km_sweeper_nephew_objects_mail { community_id time action object_type object_name object_id object_author nephew_list object_list } {
} {
    set template "km_sweeper_nephew_$action\_$time"
    set replace_list [list \
	    ad_url [ad_url] \
	    object_type $object_type \
	    object_name $object_name \
	    object_id $object_id \
	    object_author $object_author \
	    object_list $object_list \
	    nephew_list $nephew_list]
    set result [et_process -community_id $community_id $template $replace_list]

    return [ns_set get $result body]
}


ad_proc km_uncles_descendant_objects_mail { community_id time action object_type object_name object_id object_author nephew_list object_list } {
} {
    set template "km_sweeper_descendants_$action\_$time"
    set replace_list [list \
	    ad_url [ad_url] \
	    object_type $object_type \
	    object_name $object_name \
	    object_id $object_id \
	    object_author $object_author \
	    object_list $object_list \
	    nephew_list $nephew_list]
    set result [et_process -community_id $community_id $template $replace_list]

    return [ns_set get $result body]
}


ad_proc km_sweeper_object_with_children_mail { community_id time action object_type object_name object_id days_unchanged object_list } {
} {
    set template "km_sweeper_parent_$action\_$time"
    set replace_list [list \
	    ad_url [ad_url] \
	    object_type $object_type \
	    object_name $object_name \
	    object_id $object_id \
	    object_list $object_list]
    if {[string equal $time "outdated"]} {
	lappend replace_list days_unchanged $days_unchanged
    }
    set result [et_process -community_id $community_id $template $replace_list]

    return [ns_set get $result body]
}


ad_proc km_sweeper_child_object_mail { community_id time action object_type object_name object_id object_author days_unchanged object_list } {
} {
    set template "km_sweeper_child_$action\_$time"
    set replace_list [list \
	    ad_url [ad_url] \
	    object_type $object_type \
	    object_name $object_name \
	    object_id $object_id \
	    object_author $object_author \
	    object_list $object_list]
    if {[string equal $time "outdated"]} {
	lappend replace_list days_unchanged $days_unchanged
    }
    set result [et_process -community_id $community_id $template $replace_list]

    return [ns_set get $result body]
}


ad_proc km_sweeper_single_objects_mail { community_id time action object_type object_name object_id days_unchanged } {
} {
    set template "km_sweeper_single_$action\_$time"
    set replace_list [list \
	    ad_url [ad_url] \
	    object_type $object_type \
	    object_name $object_name \
	    object_id $object_id]
    if {[string equal $time "outdated"]} {
	lappend replace_list days_unchanged $days_unchanged
    }
    set result [et_process -community_id $community_id $template $replace_list]

    return [ns_set get $result body]
}


ad_proc km_sweeper_single_objects_warning_mail { time action object_type object_name object_id days_unchanged days_untill_change } {
} {
    set template "km_warn_sweeper_$action\_$time"
    set replace_list [list \
	    ad_url [ad_url] \
	    object_type $object_type \
	    object_name $object_name \
	    object_id $object_id \
	    days_untill_change $days_untill_change]
    if {[string equal $time "outdated"]} {
	lappend replace_list days_unchanged $days_unchanged
    }
    set result [et_process -community_id 0 $template $replace_list]

    return [ns_set get $result body]
}


ad_proc km_copy_object { object_id target_id } {
    copies the object object_id to the new object target_id.
    copies only active questions and no parent-child/uncle-nephew questions
} {
    set user_id [ad_conn user_id]
    set package_id [ad_conn package_id]
    set ip_addr [ns_conn peeraddr]

    db_1row get_object_name {
	select one_line_description as object_name, object_type_id
	from sn_objects
	where object_id = :object_id
    }
    set object_type [km_static object_type_pretty_name $object_type_id]
    if [empty_string_p $object_name] {
	set object_name "Unnamed $object_type"
    }
    set msg "Copied from $object_type \"$object_name\""

    db_transaction {
	set target_id [db_exec_plsql create_object {
	    begin
	    :1 := object.copy_object (
	       v_object_id     => :object_id,
	       v_target_id     => :target_id,
	       v_creation_user => :user_id,
	       v_creation_ip   => :ip_addr,
	       v_context_id    => :package_id
	    );
	    end;
	}]

	db_exec_plsql grant_owner_permissions {
	    begin
	    acs_permission.grant_permission(:target_id, :user_id, 'read');
	    acs_permission.grant_permission(:target_id, :user_id, 'write');
	    acs_permission.grant_permission(:target_id, :user_id, 'delete');
	    end;
	}

	db_dml add_audit_entry {
	    insert into sn_audit_table
	        (object_id, question_id, last_modified, last_modifying_user_id,
	         content)
	    values
	        (:target_id, null, sysdate, :user_id, :msg)
	}
    }

    return $target_id
}

ad_proc km_clean_sweeper_tmp_table { } {
    Deletes all old data from sweepers temp table in the event that the sweeper
    didn't finish.
} {
    db_dml km_sweeper_cleanup {
	delete from km_sweeper
    }
    db_dml km_sweeper_publisher_cleanup {
	delete from km_sweeper_publisher
    }
}
